use hello::{
    ThreadPool,
};
use std::{
    fs,
    io::{prelude::*, BufReader},
    net::{TcpListener, TcpStream},
    collections::HashMap,
};
use config::Config;

//CONSTS
static ROOT_FOLDER: &str = "./html";
static CONF_FILE: &str = "websrv_conf";


fn main() {
    //building config
    let cfg = Config::builder()
        .add_source(config::File::with_name(CONF_FILE))
        .build()
        .unwrap();

    let cfg = cfg.try_deserialize::<HashMap<String, String>>().expect("Could not deserialize configuration");
    println!("Starting with configuration: ");
    for conf in &cfg {
        println!("{:?}", conf);
    }

    let nbr_of_threads: usize = cfg
        .get("threads")
        .expect("Threads not found in config")
        .parse()
        .unwrap();

    let default_bind_addr = String::from("0.0.0.0:7878");
    let bind_addr = cfg
        .get("bind_addr")
        .unwrap_or(&default_bind_addr)
        .to_string();

    let listener = TcpListener::bind(bind_addr).unwrap();

    let pool = ThreadPool::new(nbr_of_threads);

    for stream in listener.incoming() {
        let stream = stream.unwrap();

        pool.execute(move || {
           handle_connection(stream);
        });
    }

    println!("Shutting down.");
}

fn search_page(page: &String) -> Option<Vec<u8>> {
    println!("Looking for {page} in {ROOT_FOLDER}");
    //regular page
    if !page.contains('.') {
        let possible_pagenames = vec![format!("{page}.html"), format!("{page}.htm")];
        for p in possible_pagenames {
            match fs::read(format!("{ROOT_FOLDER}/{p}")) {
                Ok(c) => return Some(c),
                Err(_) => (),
            }
        }
        return None;
    //css, jpg, etc
    } else {
        match fs::read(format!("{ROOT_FOLDER}/{page}")) {
            Ok(c) => return Some(c),
            Err(_) => return None,
        };
    }
}
        
///Analyze the request and answer it
///Returns the complete answer, as a vector of bytes
fn analyze_request_line(line: Vec<&str>) -> Vec<u8> {
    let _req = String::from(line[0]);
    let mut page = String::from(line[1]);
    let proto = String::from(line[2]);
    let mut ret_code = "404 NOT FOUND";

    //Removing the leading / from the page if need be
    if &page != "/" {
        page.remove(0);
    } else {
        page = String::from("index");
    }
    
    let mut content = match search_page(&page) {
        Some(c) => {
            *&mut ret_code = "200 OK";
            c
        },
        None => {
            fs::read("./html/err_pages/404.html").unwrap_or(String::from("404 NOT FOUND GET LOST").as_bytes().to_vec())
        },
    };

    let length = content.len();

    //Try to read content
    match String::from_utf8(content.clone()) {
        Ok(c) => {
            let length = c.len();
            let returned_line = format!(
                "{proto} {ret_code}\r\nContent-Length: {length}\r\n\r\n{c}"
                );
            let returned_line = returned_line.as_bytes().to_vec();
            return returned_line;
        },
        Err(_e) => {
            println!("Got {_e} since we are sending a picture");
        },
    };
    
    let img_format = if page.ends_with(".jpg") || page.ends_with(".jpeg") {
        "jpg"
    } else {
        "png"
    };

    let returned_line = format!(
        "{proto} {ret_code}\r\ncontent-type: image/{img_format}\r\ncontent-length: {length}\r\n\r\n"
        );
    let mut returned_line = returned_line.as_bytes().to_vec();
    returned_line.append(&mut content);
    return returned_line;

}

fn handle_connection(mut stream: TcpStream) {
    let buf_reader = BufReader::new(&mut stream);
    //for i in buf_reader.lines() {
    //    println!("{:?}", i);
    //}
    let mut req_lines = buf_reader.lines();
    let request = match req_lines.next() {
        Some(c) => c,
        None => Ok(String::from("RIEN")),
    };
    let request = request.unwrap();
    println!("!!{:?}", request);
    //Some(Ok("GET /crabe HTTP/1.1"))


    if &request != "RIEN" {
        let request_analyze: Vec<&str> = request.split_whitespace().collect();
        let ret_line = analyze_request_line(request_analyze);

        let payload: &[u8] = &ret_line; 
        stream.write_all(payload).unwrap();
    }
}


